#!/bin/bash -e

unset KERNEL_REL
unset KERNEL_PATCH
unset RC_PATCH
unset BUILD
unset CC
unset LINUX_GIT
unset STABLE_PATCH
unset BISECT

. version.sh

DIR=$PWD

mkdir -p ${DIR}/deploy/

function git_kernel {
	echo "Cleaning Up"
	rm -rfd ${DIR}/KERNEL || true
	mkdir -p ${DIR}/KERNEL
	git clone --shared ${LINUX_GIT} ${DIR}/KERNEL
	cd ${DIR}/KERNEL
if [ "${KERNEL_PATCH}" ] ; then
	git checkout v${KERNEL_PATCH} -b v${KERNEL_PATCH}-${BUILD}
else
	git checkout v${KERNEL_REL} -b v${KERNEL_REL}-${BUILD}
fi
	cd ${DIR}/
}

function patch_kernel {
	cd ${DIR}/KERNEL
	export DIR KERNEL_REL GIT BRANCH REL BOARD BISECT
	/bin/bash -e ${DIR}/patch.sh
if [ "${STABLE_PATCH}" ] ; then
	sed -i 's/EXTRAVERSION = '$STABLE_PATCH'/EXTRAVERSION = '$STABLE_PATCH'-'$BUILD'/g' ${DIR}/KERNEL/Makefile
	git add .
	git commit -a -m '2.6-dev '$KERNEL_PATCH'-'$BUILD' patchset'
else if [ "${RC_PATCH}" ] ; then
	sed -i 's/EXTRAVERSION = '$RC_PATCH'/EXTRAVERSION = '$RC_PATCH'-'$BUILD'/g' ${DIR}/KERNEL/Makefile
	git add .
	git commit -a -m '2.6-dev '$KERNEL_PATCH'-'$BUILD' patchset'
else
	sed -i 's/EXTRAVERSION = /EXTRAVERSION = -'$BUILD'/g' ${DIR}/KERNEL/Makefile
	git add .
	git commit -a -m '2.6-dev '$KERNEL_REL'-'$BUILD' patchset'
fi
fi
	cd ${DIR}/
}

function copy_defconfig {
	cd ${DIR}/KERNEL/
	make ARCH=arm CROSS_COMPILE=${CC} distclean
	cp ${DIR}/patches/defconfig .config
	cd ${DIR}/
}

function make_menuconfig {
	cd ${DIR}/KERNEL/
	make ARCH=arm CROSS_COMPILE=${CC} menuconfig
	cp .config ${DIR}/patches/defconfig
	cd ${DIR}/
}

function make_deb {
	cd ${DIR}/KERNEL/
	fakeroot make ARCH=arm CROSS_COMPILE=${CC} KDEB_PKGVERSION=${BUILDREV}${DISTRO} deb-pkg
	mv ${DIR}/*.deb ${DIR}/deploy/
	cd ${DIR}
}

if [ -e ${DIR}/system.sh ]; then
	. system.sh
if [ "-${LINUX_GIT}-" != "--" ]; then

	git_kernel
	patch_kernel
	copy_defconfig
	#make_menuconfig
	make_deb
else
	echo "The LINUX_GIT variable is not definted in system.sh"
	echo "Follow the git clone directions in system.sh.sample"
	echo "and make sure to remove the comment # from LINUX_GIT"
	echo "gedit system.sh"
fi
else
	echo "Missing system.sh, please copy system.sh.sample to system.sh and edit as needed"
	echo "cp system.sh.sample system.sh"
	echo "gedit system.sh"
fi

