#!/bin/bash -e

unset KERNEL_REL
unset KERNEL_PATCH
unset RC_PATCH
unset BUILD
unset CC
unset LINUX_GIT
unset STABLE_PATCH
unset BISECT

. version.sh

CCACHE=ccache
DIR=$PWD

mkdir -p ${DIR}/deploy/

function git_kernel {
	echo "Cleaning Up"
	rm -rfd ${DIR}/KERNEL || true
	mkdir -p ${DIR}/KERNEL
	cd ${LINUX_GIT}/
	git fetch
	cd -
	git clone --shared ${LINUX_GIT} ${DIR}/KERNEL
	cd ${DIR}/KERNEL
if [ "${KERNEL_PATCH}" ] ; then
	git checkout v${KERNEL_PATCH} -b v${KERNEL_PATCH}-${BUILD}
else
	git checkout v${KERNEL_REL} -b v${KERNEL_REL}-${BUILD}
fi
	cd ${DIR}/
}

function patch_kernel {
	cd ${DIR}/KERNEL
	export DIR KERNEL_REL GIT BRANCH REL BOARD BISECT
	/bin/bash -e ${DIR}/patch.sh
if [ "${STABLE_PATCH}" ] ; then
	sed -i 's/EXTRAVERSION = '$STABLE_PATCH'/EXTRAVERSION = '$STABLE_PATCH'-'$BUILD'/g' ${DIR}/KERNEL/Makefile
	git add .
	git commit -a -m '2.6-dev '$KERNEL_PATCH'-'$BUILD' patchset'
else if [ "${RC_PATCH}" ] ; then
	sed -i 's/EXTRAVERSION = '$RC_PATCH'/EXTRAVERSION = '$RC_PATCH'-'$BUILD'/g' ${DIR}/KERNEL/Makefile
	git add .
	git commit -a -m '2.6-dev '$KERNEL_PATCH'-'$BUILD' patchset'
else
	sed -i 's/EXTRAVERSION = /EXTRAVERSION = -'$BUILD'/g' ${DIR}/KERNEL/Makefile
	git add .
	git commit -a -m '2.6-dev '$KERNEL_REL'-'$BUILD' patchset'
fi
fi
	cd ${DIR}/
}

function dspbridge {
	cd ${DIR}/KERNEL/
	echo "Merging dspbridge"
	git remote add -t dspbridge tmlind-omap-dspbridge git://git.kernel.org/pub/scm/linux/kernel/git/tmlind/linux-omap-2.6.git
	git pull tmlind-omap-dspbridge dspbridge
	echo "fix conflicts"
	gedit arch/arm/mach-omap2/io.c
	git commit -a -m 'dsp merge'
	#git apply ${DIR}/patches/dsp/dspbridge-merge.diff
	#git add .
	#git commit -a -m 'dsp next'
	git apply ${DIR}/patches/dsp/dspbridge-merge.diff
	cd ${DIR}/
}

function use_angstrom_defconfig {
	cd ${DIR}/KERNEL/
	make ARCH=arm CROSS_COMPILE=${CC} distclean
	cp ${DIR}/patches/angstrom/beagleboard/defconfig .config
	cd ${DIR}/
}

function copy_defconfig {
	cd ${DIR}/KERNEL/
	make ARCH=arm CROSS_COMPILE=${CC} distclean
	cp ${DIR}/patches/defconfig .config
	cd ${DIR}/
}

function make_menuconfig {
	cd ${DIR}/KERNEL/
	make ARCH=arm CROSS_COMPILE=${CC} menuconfig
	cp .config ${DIR}/patches/defconfig
	cd ${DIR}/
}

function make_uImage {
	cd ${DIR}/KERNEL/
	time make -j2 ARCH=arm CROSS_COMPILE="${CCACHE} ${CC}" uImage
	KERNEL_UTS=$(cat ${DIR}/KERNEL/include/generated/utsrelease.h | awk '{print $3}' | sed 's/\"//g' )
	cp arch/arm/boot/uImage ${DIR}/deploy/${KERNEL_UTS}.uImage
	cd ${DIR}
}

function make_modules {
	cd ${DIR}/KERNEL/
	time make -j2 ARCH=arm CROSS_COMPILE="${CCACHE} ${CC}" modules
	rm -rfd ${DIR}/deploy/mod &> /dev/null || true
	mkdir -p ${DIR}/deploy/mod
	make ARCH=arm CROSS_COMPILE=${CC} modules_install INSTALL_MOD_PATH=${DIR}/deploy/mod
	cd ${DIR}/deploy/mod
	tar czf ../${KERNEL_UTS}-modules.tar.gz *
	cd ${DIR}
}

if [ -e ${DIR}/system.sh ]; then
	. system.sh
if [ "-${LINUX_GIT}-" != "--" ]; then

	git_kernel
	patch_kernel
	#dspbridge
	#use_angstrom_defconfig
	copy_defconfig
	make_menuconfig
	make_uImage
	make_modules
else
	echo "The LINUX_GIT variable is not definted in system.sh"
	echo "Follow the git clone directions in system.sh.sample"
	echo "and make sure to remove the comment # from LINUX_GIT"
	echo "gedit system.sh"
fi
else
	echo "Missing system.sh, please copy system.sh.sample to system.sh and edit as needed"
	echo "cp system.sh.sample system.sh"
	echo "gedit system.sh"
fi

