#!/bin/bash -e
#DSP Link
#Requirements: http://cgit.openembedded.org/cgit.cgi/openembedded/tree/recipes/ti/README

unset CC

. version.sh

echo ""
echo "NO SUPPORT, just a proof of concept, dsplinkk.ko build isn't completing..."
echo ""

DIR=$PWD

CODEC_ENGINE=codec_engine_2_24_01.tar.gz
DSPLINK_TAR=dsplink_linux_1_63.tar.gz

BIOS_VER=5_33_04
BIOS_FILE_VER=5.33.04
BIOS=bios_setuplinux_${BIOS_VER}.bin
BIOS_DIR=bios_${BIOS_VER}
BIOS_FILE=DSP_BIOS_${BIOS_FILE_VER}_Components

CGVERSION=6.1.9
CGTOOLS_BIN=ti_cgt_c6000_${CGVERSION}_setup_linux_x86.bin

XDCVER=3_15_00_50
XDCTOOLS_BIN=xdctools_setuplinux_${XDCVER}.bin

DSPLINKPLATFORM=OMAP3530
DSPLINKDSPCFG=OMAP3530SHMEM
DSPLINKGPPOS=OMAPLSP

if [ -e ${DIR}/system.sh ]; then
	. system.sh

if [ -e ${DIR}/dl/${CODEC_ENGINE} ]; then
  echo "Code Engine found"
  BUILD=C
else
  echo "Code Engine not found: copy ${CODEC_ENGINE} to ./dl : http://cgit.openembedded.org/cgit.cgi/openembedded/tree/recipes/ti/README"
fi

if [ -e ${DIR}/dl/${BIOS} ]; then
  echo "${BIOS} found"
  BUILD+=B
  if [ -e  ${DIR}/dl/${BIOS_DIR}/${BIOS_FILE} ]; then
    echo "Extracted ${BIOS} found"
  else
    cd ${DIR}/dl/
    echo "${BIOS} needs to be executable"
    sudo chmod +x ./${BIOS}
    echo "running ${BIOS}, accept all defaults and agree to the license"
    ./${BIOS} --mode console --prefix ${DIR}/dl/${BIOS_DIR}
    cd ${DIR}
  fi
else
  echo ""
  echo "${BIOS} not found"
  echo "Download ${BIOS}"
  echo "DL From: http://software-dl.ti.com/dsps/dsps_registered_sw/sdo_sb/targetcontent/bios/bios_5_33/bios_5_33_04/index_external.html" 
  echo "Copy to: ${DIR}/dl"
  echo ""
fi

if [ -e ${DIR}/dl/${CGTOOLS_BIN} ]; then
  echo "${CGTOOLS_BIN} found"
  BUILD+=C
  if [ -e  ${DIR}/dl/TI_CGT_C6000_${CGVERSION}/uninstall_cgt_c6000.bin ]; then
    echo "Extracted ${CGTOOLS_BIN} found"
  else
    cd ${DIR}/dl/
    echo "${CGTOOLS_BIN} needs to be executable"
    sudo chmod +x ./${CGTOOLS_BIN}
    echo "running ${CGTOOLS_BIN}, accept all defaults and agree to the license"
    ./${CGTOOLS_BIN} --mode console --prefix ${DIR}/dl/TI_CGT_C6000_${CGVERSION}
    cd ${DIR}
  fi
else
  echo ""
  echo "${CGTOOLS_BIN} not found"
  echo "Download ${CGTOOLS_BIN}"
  echo "DL From: https://www-a.ti.com/downloads/sds_support/TICodegenerationTools/download.htm" 
  echo "Copy to: ${DIR}/dl"
  echo ""
fi

if [ -e ${DIR}/dl/${DSPLINK_TAR} ]; then
  echo "DSP link found"
  BUILD+=D
else
  echo "DSP link not found: copy ${DSPLINK_TAR} to ./dl : http://cgit.openembedded.org/cgit.cgi/openembedded/tree/recipes/ti/README"
fi

if [ -e ${DIR}/dl/${XDCTOOLS_BIN} ]; then
  echo "${XDCTOOLS_BIN} found"
  BUILD+=X
  if [ -e  ${DIR}/dl/xdctools_${XDCVER}/xdctools_${XDCVER}_release_notes.html ]; then
    echo "Extracted ${XDCTOOLS_BIN} found"
  else
    cd ${DIR}/dl/
    echo "${XDCTOOLS_BIN} needs to be executable"
    sudo chmod +x ./${XDCTOOLS_BIN}
    echo "running ${XDCTOOLS_BIN}, accept all defaults and agree to the license"
    ./${XDCTOOLS_BIN} --mode console --prefix ${DIR}/dl/xdctools_${XDCVER}
    cd ${DIR}
  fi
else
  echo ""
  echo "${XDCTOOLS_BIN} not found"
  echo "Download ${XDCTOOLS_BIN}"
  echo "DL From: http://software-dl.ti.com/dsps/dsps_registered_sw/sdo_sb/targetcontent/" 
  echo "Copy to: ${DIR}/dl"
  echo ""
fi

echo "Sanity Check"
echo $BUILD

if [ "$BUILD" = "CBCDX" ]; then

if [ -e ${DIR}/KERNEL/arch/arm/boot/uImage ]; then
{
  mkdir -p ${DIR}/deploy/dsp

  export DSPLINK=${DIR}/dsplink_linux_1_63/dsplink
  echo "Cleaning up"
  rm -rfd ${DIR}/dsp* || true
  echo "Extracting DSPLINK"
  tar xf ${DIR}/dl/${DSPLINK_TAR}
  cd ${DSPLINK}
  sed -i /page.h/d ${DSPLINK}/dsplink/gpp/src/api/Linux/drv_api.c
  perl config/bin/dsplinkcfg.pl --platform=${DSPLINKPLATFORM} --nodsp=1 --dspcfg_0=${DSPLINKDSPCFG} --dspos_0=DSPBIOS5XX --gppos=${DSPLINKGPPOS} --comps=ponslrm
  sed -i 's/OMAP3530/${DSPLINKPLATFORM}/g' ${DSPLINK}/gpp/src/Makefile

  unset CFLAGS CPPFLAGS CXXFLAGS LDFLAGS

TOOLPATH=/OE/angstrom-dev/cross/armv7a/bin

TARGET_PREFIX=arm-angstrom-linux-gnueabi-
KERNEL_CC="${TOOLPATH}/${TARGET_PREFIX}gcc"
KERNEL_AR="${TOOLPATH}/${TARGET_PREFIX}ar"
KERNEL_LD="${TOOLPATH}/${TARGET_PREFIX}ld"
STAGING_KERNEL_DIR="${DIR}/KERNEL/"

    echo ""
    echo "Building the gpp user space library"
    cd ${DSPLINK}/gpp/src/api
    make \
      CROSS_COMPILE="${TARGET_PREFIX}" \
      CC="${KERNEL_CC}" \
      AR="${KERNEL_AR}" \
      LD="${KERNEL_LD}" \
      COMPILER="${KERNEL_CC}" \
      ARCHIVER="${KERNEL_AR}" \
      KERNEL_DIR="${STAGING_KERNEL_DIR}" \
      clean all > ${DIR}/gpp-user-space.log

    echo ""
    echo "Building the gpp kernel space (debug and release)"
    cd ${DSPLINK}/gpp/src
    make \
      OBJDUMP="${TARGET_PREFIX}objdump" \
      CROSS_COMPILE="${TARGET_PREFIX}" \
      CC="${KERNEL_CC}" \
      AR="${KERNEL_AR}" \
      LD="${KERNEL_LD}" \
      COMPILER="${KERNEL_CC}" \
      ARCHIVER="${KERNEL_AR}" \
      KERNEL_DIR="${STAGING_KERNEL_DIR}" \
      clean all > ${DIR}/gpp-kernel-space.log

CROSS_DIR=/OE/angstrom-dev/cross/armv7a
TARGET_SYS=arm-angstrom-linux-gnueabi

    echo ""
    echo "Building the gpp samples"
    cd ${DSPLINK}/gpp/src/samples
    make \
      BASE_TOOLCHAIN="${CROSS_DIR}" \
      BASE_CGTOOLS="${BASE_TOOLCHAIN}/bin" \
      OSINC_PLATFORM="${CROSS_DIR}/lib/gcc/${TARGET_SYS}/$(${TOOLPATH}/${TARGET_PREFIX}gcc -dumpversion)/include" \
      OSINC_TARGET="${BASE_TOOLCHAIN}/target/usr/include" \
      CROSS_COMPILE="${TARGET_PREFIX}" \
      CC="${KERNEL_CC}" \
      AR="${KERNEL_AR}" \
      LD="${KERNEL_LD}" \
      COMPILER="${KERNEL_CC}" \
      LINKER="${KERNEL_CC}" \
      ARCHIVER="${KERNEL_AR}" \
      KERNEL_DIR="${STAGING_KERNEL_DIR}" \
      clean all > ${DIR}/gpp-samples.log

STAGING_TI_CGT6x_DIR=${DIR}/dl/TI_CGT_C6000_${CGVERSION}
STAGING_TI_DSPBIOS_DIR=${DIR}/dl/bios_${BIOS_VER}
STAGING_TI_XDCTOOL_INSTALL_DIR=${DIR}/dl/xdctools_${XDCVER}

#weird bug, moveing to a non '.' directory...
rm -rfd /tmp/bios &> /dev/null
mkdir -p /tmp/bios &> /dev/null
cp -pPrf ${STAGING_TI_DSPBIOS_DIR}/* /tmp/bios/
STAGING_TI_DSPBIOS_DIR=/tmp/bios

    echo ""
    echo "Building the dsp library (debug and release)"
    cd ${DSPLINK}/dsp/src
    make \
      BASE_CGTOOLS="${STAGING_TI_CGT6x_DIR}" \
      BASE_SABIOS="${STAGING_TI_DSPBIOS_DIR}" \
      clean all > ${DIR}/dsp-library.log

    echo ""
    echo "Building the dsp samples (debug and release)"
    cd ${DSPLINK}/dsp/src/samples
    make \
      BASE_CGTOOLS="${STAGING_TI_CGT6x_DIR}" \
      BASE_SABIOS="${STAGING_TI_DSPBIOS_DIR}" \
      clean all > ${DIR}/dsp-samples.log

cp ${DSPLINK}/gpp/export/BIN/Linux/${DSPLINKPLATFORM}/RELEASE/dsplinkk.ko ${DIR}/deploy/dsp/
cp ${DSPLINK}/gpp/export/BIN/Linux/${DSPLINKPLATFORM}/RELEASE/dsplink.lib ${DIR}/deploy/dsp/
mkdir -p ${DIR}/deploy/dsp/apps
cp ${DSPLINK}/gpp/export/BIN/Linux/${DSPLINKPLATFORM}/RELEASE/*gpp ${DIR}/deploy/dsp/apps/

    for i in $(find ${DSPLINK}/dsp/BUILD/ -name "*.out") ; do
        cp ${i} ${DIR}/deploy/dsp/apps/
    done

cd ${DIR}

}
else
{

echo "run build_kernel.sh first"

}
fi
fi
else
	echo "Missing system.sh, please copy system.sh.sample to system.sh and edit as needed"
	echo "cp system.sh.sample system.sh"
	echo "gedit system.sh"
fi

